-- ========================================================================== --
--   Mandelbrot.as                                                            --
-- -------------------------------------------------------------------------- --
--   Copyright (c) 2012 Laurens Rodriguez Oscanoa.                            --
--   This code is licensed under the MIT license:                             --
--   http://www.opensource.org/licenses/mit-license.php                       --
-- -------------------------------------------------------------------------- --

local INIT_STEPS = 256
local INCREMENT_STEPS = 64

local MINIMUM_COLOR_STEP = 5
local COLOR_RANGE = 64
local COLOR_STEP = 6

local COLOR_BLUE   = 1
local COLOR_RED    = 2
local COLOR_GREEN  = 3
local COLOR_GRAY   = 4
local COLOR_RANDOM = 5
local COLOR_PALETTE = 6

local MINIMUM_CELL_SIZE = 9e-15

-- The fractal zone and the zoom zone are divided in cells.
local FRACTAL_GRID_CELLS = 8
local ZOOM_GRID_CELLS = 4

-------------------------------------------------------------------------------
local m_width  -- canvas width
local m_height -- canvas height

local m_canvas

local m_bitmapBack
local m_bitmapFront

local m_colors -- palette of colors
local m_colorSteps

-- Number of iterations for computing a pixel of the fractal.
local m_steps

-- Fractal zone coordinates
local m_fractalZoneX1
local m_fractalZoneY1
local m_fractalZoneX2
local m_fractalZoneY2

-- Zoom zone position on screen
local m_zoomX1
local m_zoomY1
local m_zoomX2
local m_zoomY2

-- Zoom grid cell position
local m_zoomGridX
local m_zoomGridY

local m_zoomIsVisible

-- Fractal zones history.
-- Each element is an array with the zone coordinates: [x1, y1, x2, y2]
local m_zones

-- Some nice palette colors
local m_palette

-------------------------------------------------------------------------------
function drawZoomZone()

    local cellWidth = m_width / FRACTAL_GRID_CELLS
    local cellHeight = m_height / FRACTAL_GRID_CELLS

    local x1 = cellWidth * (m_zoomGridX) - 1
    local x2 = cellWidth * (m_zoomGridX + ZOOM_GRID_CELLS) - 1

    local y1 = cellHeight * (FRACTAL_GRID_CELLS - m_zoomGridY) - 1
    local y2 = cellHeight * (FRACTAL_GRID_CELLS - m_zoomGridY - ZOOM_GRID_CELLS) - 1

    if (m_zoomIsVisible) then
        -- Erase previous zoom grid.
        m_bitmapFront:clear()
    end
    love.graphics.setCanvas(m_bitmapFront)
    love.graphics.setColor(255, 255, 255)
    love.graphics.rectangle("line", x1, y1, x2 - x1, y2 - y1)
    love.graphics.setCanvas()

    m_zoomX1 = x1
    m_zoomY1 = y2
    m_zoomX2 = x2
    m_zoomY2 = y1
end

function toggleZoomZone()
    if (not m_zoomIsVisible) then
        drawZoomZone()
        m_zoomIsVisible = true
    else
        m_bitmapFront:clear(0, 0, 0, 0)
        m_zoomIsVisible = false
    end
end

function drawMandelbrot(initializeZone)

    if (m_zoomIsVisible) then
        toggleZoomZone()
    end
    if (initializeZone) then
        m_fractalZoneX1 = -2.5
        m_fractalZoneY1 = -1.2
        m_fractalZoneX2 = 0.7
        m_fractalZoneY2 = 1.2
    end

    local dx = (m_fractalZoneX2 - m_fractalZoneX1) / (m_width - 1)
    local dy = (m_fractalZoneY2 - m_fractalZoneY1) / (m_height - 1)

    love.graphics.setCanvas(m_bitmapBack)
    m_bitmapBack:clear(0, 0, 0)

	local px, py
	local steps
	local fx, fy
	local temp

    -- Draw fractal zone
    for x = 0, m_width - 1 do
        for y = 0, m_height - 1 do
            -- Point in fractal zone
            px = m_fractalZoneX1 + x * dx
            py = m_fractalZoneY2 - y * dy

            -- Iterate fractal computation.
            steps = 0
            fx = 0.0
            fy = 0.0

            while (true) do
                -- Mandelbrot recurrence:
                -- ---------------------
                -- F(n+1) = F(n)*F(n) + (px + i*py)
                temp = fx*fx - fy*fy + px
                fy = 2*fx*fy + py
                fx = temp

                steps = steps + 1

                -- F(z) belongs to Mandelbrot set if |F(z)| < 2
                -- We give up if we passed the limit of number of iterations.
                if ((steps >= m_steps) or (fx*fx + fy*fy >= 4.0)) then
                    break
                end
            end

            if (steps < m_steps) then
                -- We found that: |F(z)| >= 2 (the point doesn't belong to the Mandelbrot set)
                local indexColor = (steps - 1) % 28 + 1
                if (indexColor > 15) then
                    indexColor = 30 - indexColor
                end
				indexColor = indexColor + 1
                love.graphics.setColor(m_colors[indexColor][1], m_colors[indexColor][2], m_colors[indexColor][3])
                love.graphics.point(x, y)
            else
                -- We suspect this point belongs to the Mandelbrot set.
                -- We can't be sure because the point can belong to the set if the number
                -- of iteration steps is increased. Leave this pixel in black.
            end
        end
    end
    love.graphics.setCanvas()
    love.graphics.setColor(255, 255, 255)
end

function redraw()
    if (m_zoomIsVisible) then
        -- If the zoom grid is visible we want to redraw the zoom zone.
        local dx = (m_fractalZoneX2 - m_fractalZoneX1) / FRACTAL_GRID_CELLS
        local dy = (m_fractalZoneY2 - m_fractalZoneY1) / FRACTAL_GRID_CELLS

        if ((dx >= MINIMUM_CELL_SIZE) and (dy >= MINIMUM_CELL_SIZE)) then

            -- Save actual zone position
            table.insert(m_zones, {m_fractalZoneX1, m_fractalZoneY1, m_fractalZoneX2, m_fractalZoneY2})

            -- Set new drawing zone. (x1,y1) bottom-left corner, (x2,y2) up-right corner
            m_fractalZoneX1 = m_fractalZoneX1 + (m_zoomGridX * dx)
            m_fractalZoneY1 = m_fractalZoneY1 + (m_zoomGridY * dy)
            m_fractalZoneX2 = m_fractalZoneX1 + ZOOM_GRID_CELLS * dx
            m_fractalZoneY2 = m_fractalZoneY1 + ZOOM_GRID_CELLS * dy

            drawMandelbrot()
        end
    else
        toggleZoomZone()
    end
end

function increaseSteps(increment)
    if (m_steps + increment > 0) then
        m_steps = m_steps + increment
        drawMandelbrot()
    end
end

function zoomOut()
    if (m_zoomIsVisible) then
        toggleZoomZone()
    end
    if (#m_zones > 0) then
        local zone = table.remove(m_zones)
        m_fractalZoneX1 = zone[1]
        m_fractalZoneY1 = zone[2]
        m_fractalZoneX2 = zone[3]
        m_fractalZoneY2 = zone[4]
    else
        local dx = (m_fractalZoneX2 - m_fractalZoneX1)/2
        local dy = (m_fractalZoneY2 - m_fractalZoneY1)/2
        m_fractalZoneX1 = m_fractalZoneX1 - dx
        m_fractalZoneY1 = m_fractalZoneY1 - dy
        m_fractalZoneX2 = m_fractalZoneX2 + dx
        m_fractalZoneY2 = m_fractalZoneX2 + dy
    end
    drawMandelbrot()
end

function setColors(opc)
    local k
    m_colors = {}

    if (opc == COLOR_GRAY) then
        for k = 1, 16 do
            table.insert(m_colors, {4*m_colorSteps[k][2], 4*m_colorSteps[k][2], 4*m_colorSteps[k][2]})
        end
    elseif (opc == COLOR_BLUE) then
        for k = 1, 16 do
            table.insert(m_colors, {4*m_colorSteps[k][1], 4*m_colorSteps[k][2], 4*m_colorSteps[k][3]})
        end
    elseif (opc == COLOR_RED) then
        for k = 1, 16 do
            table.insert(m_colors, {4*m_colorSteps[k][3], 13 * k, 3*m_colorSteps[k][1]})
        end
    elseif (opc == COLOR_GREEN) then
        for k = 1, 16 do
            table.insert(m_colors, {14 * k, 4*m_colorSteps[k][3], 4*m_colorSteps[k][1]})
        end
    elseif (opc == COLOR_RANDOM) then
        local c1, c2, c3
        repeat
            c1 = math.random(COLOR_STEP) - 1
            c2 = math.random(COLOR_STEP) - 1
            c3 = math.random(COLOR_STEP) - 1

        until ((c1 + c2 + c3) >= MINIMUM_COLOR_STEP)

        table.insert(m_colors, {0, 0, 0}) -- base color is black

        local a1 = math.random(COLOR_RANGE) - 1
        local a2 = math.random(COLOR_RANGE) - 1
        local a3 = math.random(COLOR_RANGE) - 1

        -- Fill color table.
        for k = 1, 15 do
            local t1 = (a1 + (k-1) * c1) % (COLOR_RANGE * 2)
            local t2 = (a2 + (k-1) * c2) % (COLOR_RANGE * 2)
            local t3 = (a3 + (k-1) * c3) % (COLOR_RANGE * 2)

            if (t1 >= COLOR_RANGE) then
                t1 = (COLOR_RANGE * 2) - t1 - 1
            end
            if (t2 >= COLOR_RANGE) then
                t2 = (COLOR_RANGE * 2) - t2 - 1
            end
            if (t3 >= COLOR_RANGE) then
                t3 = (COLOR_RANGE * 2) - t3 - 1
            end
            table.insert(m_colors, {4*t1, 4*t2, 4*t3})
        end
    elseif (opc == COLOR_PALETTE) then
        local palette = math.random(#m_palette)
        table.insert(m_colors, {0, 0, 0}) -- base color is black

        -- Fill color table.
        for k = 1, 15 do
            local p1 = (m_palette[palette][1] + (k-1) * m_palette[palette][2]) % (COLOR_RANGE * 2)
            local p2 = (m_palette[palette][3] + (k-1) * m_palette[palette][4]) % (COLOR_RANGE * 2)
            local p3 = (m_palette[palette][5] + (k-1) * m_palette[palette][6]) % (COLOR_RANGE * 2)

            if (p1 >= COLOR_RANGE) then
                p1 = (COLOR_RANGE * 2) - p1 - 1
            end
            if (p2 >= COLOR_RANGE) then
                p2 = (COLOR_RANGE * 2) - p2 - 1
            end
            if (p3 >= COLOR_RANGE) then
                p3 = (COLOR_RANGE * 2) - p3 - 1
            end
            table.insert(m_colors, {4*p1, 4*p2, 4*p3})
        end
    end
end

function moveZoomLeft()
    if (not m_zoomIsVisible) then
        toggleZoomZone()
    else
        if (m_zoomGridX > 0) then
            m_zoomGridX = m_zoomGridX - 1
            drawZoomZone()
        end
    end
end

function moveZoomRight()
    if (not m_zoomIsVisible) then
        toggleZoomZone()
    else
        if (m_zoomGridX < FRACTAL_GRID_CELLS - ZOOM_GRID_CELLS) then
            m_zoomGridX = m_zoomGridX + 1
            drawZoomZone()
        end
    end
end

function moveZoomUp()
    if (not m_zoomIsVisible) then
        toggleZoomZone()
    else
        if (m_zoomGridY < FRACTAL_GRID_CELLS - ZOOM_GRID_CELLS) then
            m_zoomGridY = m_zoomGridY + 1
            drawZoomZone()
        end
    end
end

function moveZoomDown()
    if (not m_zoomIsVisible) then
        toggleZoomZone()
    else
        if (m_zoomGridY > 0) then
            m_zoomGridY = m_zoomGridY - 1
            drawZoomZone()
        end
    end
end

function initializeColorData()
    m_palette = {
        {0, 4, 62, 5, 31, 3},  {7, 4, 4, 4, 42, 5},   {8, 0, 55, 4, 4, 4},   {8, 5, 8, 4, 8, 1},    {12, 4, 44, 2, 46, 3},
        {17, 4, 35, 5, 41, 4}, {20, 5, 43, 4, 57, 3}, {20, 5, 58, 5, 21, 2}, {21, 2, 35, 4, 59, 0}, {24, 4, 53, 2, 54, 3},
        {25, 2, 36, 2, 50, 2}, {25, 5, 52, 5, 0, 5},  {27, 3, 19, 3, 31, 5}, {27, 3, 35, 4, 39, 2}, {29, 5, 63, 2, 34, 2},
        {32, 5, 58, 5, 33, 2}, {33, 5, 61, 5, 34, 3}, {35, 2, 16, 5, 22, 0}, {35, 4, 2, 0, 10, 3},  {36, 4, 43, 4, 35, 2},
        {38, 4, 63, 3, 55, 5}, {39, 5, 8, 2, 48, 2},  {39, 5, 59, 3, 7, 2},  {40, 3, 6, 0, 61, 5},  {40, 5, 58, 5, 25, 2},
        {41, 2, 49, 5, 52, 3}, {41, 5, 59, 5, 0, 3},  {43, 5, 56, 3, 43, 4}, {44, 2, 11, 3, 54, 4}, {44, 4, 61, 4, 13, 2},
        {45, 3, 61, 3, 10, 1}, {45, 4, 63, 5, 6, 3},  {45, 5, 46, 5, 11, 0}, {46, 3, 5, 5, 17, 3},  {47, 1, 30, 4, 14, 0},
        {48, 5, 58, 5, 17, 2}, {48, 5, 63, 4, 6, 4},  {48, 5, 63, 5, 6, 4},  {49, 3, 17, 4, 38, 2}, {50, 2, 63, 5, 57, 3},
        {51, 5, 62, 3, 37, 0}, {53, 1, 56, 5, 13, 1}, {53, 3, 57, 2, 49, 1}, {53, 3, 56, 5, 44, 2}, {54, 4, 1, 0, 33, 3},
        {54, 5, 53, 4, 45, 2}, {55, 4, 13, 0, 4, 4},  {55, 4, 40, 5, 34, 2}, {55, 5, 57, 2, 56, 2}, {57, 2, 14, 3, 20, 0},
        {58, 1, 15, 5, 9, 2},  {58, 3, 38, 4, 13, 4}, {59, 0, 48, 5, 6, 2},  {59, 3, 3, 0, 6, 3},   {59, 5, 4, 4, 60, 0},
        {59, 5, 52, 3, 5, 0},  {60, 1, 51, 5, 0, 3},  {60, 5, 14, 2, 24, 3}, {61, 5, 42, 5, 24, 3}, {63, 4, 14, 3, 0, 5}
    }

    m_colorSteps = {
        { 0,  0,  0}, {10,  8, 23}, {13, 16, 36}, {15, 18, 41},
        {17, 21, 46}, {18, 23, 49}, {20, 26, 50}, {22, 29, 52},
        {24, 35, 55}, {22, 40, 58}, {20, 45, 60}, {17, 46, 61},
        {16, 47, 62}, {25, 52, 62}, {38, 58, 63}, {63, 63, 63}
    }
end

-------------------------------------------------------------------------------
function love.load()
    m_width = 640
    m_height = 480

    love.graphics.setPoint(1, "rough")
    love.graphics.setLine(1, "rough")

    m_bitmapBack = love.graphics.newCanvas(m_width, m_height)
    m_bitmapFront = love.graphics.newCanvas(m_width, m_height)

    m_steps = INIT_STEPS
    m_zoomIsVisible = false
    m_zones = {}

    -- Zoom zone init position
    m_zoomGridX = (FRACTAL_GRID_CELLS - ZOOM_GRID_CELLS) / 2
    m_zoomGridY = m_zoomGridX

    -- Initialize color data
    initializeColorData()
    setColors(COLOR_PALETTE)

    -- Draw Mandelbrot fractal in default zone
    drawMandelbrot(true)
end

function love.draw()
    love.graphics.draw(m_bitmapBack, 0, 0)
    love.graphics.draw(m_bitmapFront, 0, 0)
end

function love.keypressed(key, unicode)
    if ((key == "left") or (key == "a")) then
        moveZoomLeft()
    elseif ((key == "right") or (key == "d")) then
        moveZoomRight()
    elseif ((key == "down") or (key == "s")) then
        moveZoomDown()
    elseif ((key == "up") or (key == "w")) then
        moveZoomUp()
    elseif ((key == " ") or (key == "return")) then
        redraw()
    elseif (key == "f4") then
        setColors(COLOR_PALETTE)
        drawMandelbrot()
    elseif (key == "f5") then
        setColors(COLOR_RANDOM)
        drawMandelbrot()
    elseif (key == "f6") then
        setColors(COLOR_RED)
        drawMandelbrot()
    elseif (key == "f7") then
        setColors(COLOR_GREEN)
        drawMandelbrot()
    elseif (key == "f8") then
        setColors(COLOR_BLUE)
        drawMandelbrot()
    elseif (key == "f9") then
        setColors(COLOR_GRAY)
        drawMandelbrot()
    elseif (key == "backspace") then
        zoomOut()
    elseif (key == "escape") then
        love.event.push("quit")
    elseif (key == "home") then
        drawMandelbrot(true)
    elseif (key == "pageup") then
        increaseSteps(INCREMENT_STEPS)
    elseif (key == "pagedown") then
        increaseSteps(-INCREMENT_STEPS)
    else
        toggleZoomZone()
    end
end
